# coding: utf-8

'''
@author: vmmhypervisor
'''

import os
import sys
import re
import math

import golib
import golib.common as common
import golib.sysstat as sysstat
import golib.gtime as gtime
import golib.gfile as gfile
from golib.logger import logger


def setupSysPath():
    import_path_list = [
        "/usr/bin",
        "/usr/local/bin",
        "/usr/local/lib/python2.7/dist-packages",
        "/usr/local/lib/python2.7/site-packages",
        "/usr/local/lib/python2.7",
    ]

    for import_path in import_path_list:
        if not os.path.isdir(import_path):
            continue

        if import_path not in sys.path:
            sys.path.insert(0, import_path)

    logger.debug("sys.path: " + str(sys.path))

setupSysPath()

try:
    import rpy2.rinterface
    import rpy2.robjects as robj
    from rpy2.robjects.packages import importr
    from rpy2.robjects.lib import grid
    LOAD_RPY2 = True

    logger.debug("import success: rpy2")

    try:
        # install.packages("ggplot2") # ggplot2 required to install from CRAN
        import rpy2.robjects.lib.ggplot2 as ggplot2
        LOAD_GGPLOT2 = True
    except rpy2.rinterface.RRuntimeError:
        LOAD_GGPLOT2 = False
        logger.warn("import error: ggplot2 package not found")
except ImportError, e:
    LOAD_RPY2 = False
    LOAD_GGPLOT2 = False
    logger.warn("import error: rpy2 package")
    logger.warn(e)

"""
try:
    import multiprocessing
    IMPORT_MP = True
    plot_pool = multiprocessing.Pool(processes=None)
except ImportError:
    IMPORT_MP = False
    plot_pool = None
#"""

MIN_ImageWidthPixel = 720
MAX_ImageWidthPixel = 1872		# except boxplot

# Extension ---
EXT_PNG = "png"
EXT_SVG = "svg"

# Regular Expression ---
RE_PNG_EXT = "\.png$"


class PlotType:
    SCATTER_DIAGRAM = "scatter"
    LINE_CHART = "line"
    BAR_GRAPH = "Bar"
    BOXPLOT = "boxplot"
    HISTGRAM = "histogram"
    STACK_GRAPH = "stack"
    STACK_AREA_GRAPH = "stack-area"
    STACK_BAR_GRAPH = "stack-bar"
    RECT = "rect"

    class Stack:
        AREA = "area"
        BAR = "bar"


def getUniqueList(r_vector):
    fac_vector = robj.FactorVector(r_vector)
    return tuple(fac_vector.levels)


def getUniqueCount(r_vector):
    return len(getUniqueList(r_vector))


def getColorVector(color_count, vector_size):
    color_vector = robj.r.rainbow(color_count)

    while len(color_vector) < vector_size:
        color_vector += robj.r.gray(0.75)

    logger.debug(
        "color-count=%d, vector-size=%d" % (color_count, vector_size))

    return robj.r["rev"](color_vector)


def getGrayColorVector(color_count, vector_size):
    color_vector = None
    gray_range = 4
    gray_vector = None

    for count in range(int(color_count)):
        _div, mod = divmod(count, gray_range)
        gray = robj.r.gray(min(0.75, mod / float(gray_range) + 0.25))
        if gray_vector is None:
            gray_vector = gray
        else:
            gray_vector = gray_vector + gray

    if color_vector is None:
        color_vector = gray_vector
    else:
        color_vector = color_vector + gray_vector

    while len(color_vector) < vector_size:
        color_vector += robj.r.gray(0.75)

    logger.debug(
        "color-count=%d, vector-size=%d" % (color_count, vector_size))

    return robj.r["rev"](color_vector)


def getShapeVector(count):
    if count <= 0:
        raise ValueError()

    max_shape_id = 25
    loop_count = int(math.ceil(count / float(max_shape_id)))
    shape_vector = robj.IntVector(range(max_shape_id) * loop_count)

    logger.debug("shape values: " + str(shape_vector))

    return shape_vector


def RVectorToList(r_vector):
    fac_vector = robj.FactorVector(r_vector)

    #[fac_vector.levels[factor] for factor in list(fac_vector)]
    factor_tuple = tuple(fac_vector.levels)
    return [factor_tuple[factor - 1] for factor in list(fac_vector)]


# R data frame ---

def isRDataFrame(value):
    return common.isNotNone(value) and isinstance(value, RDataFrame)


class RDataFrame(object):
    __DEBUG_PLOT_DATAFRAME = True

    @property
    def data_frame(self):
        return self.__data_frame

    def __init__(self, data_frame=None):
        self.__data_frame = data_frame
        self.__dict_name_datalist = {}

    def printDataframe(self, n=-1):
        min_len = n
        if min_len < 0:
            min_len = sys.maxint

        for colname in self.data_frame.colnames:
            col_len = len(self.data_frame.rx2(colname))
            min_len = min(min_len, col_len)
            print colname, col_len

        print self.data_frame.rx(robj.IntVector((range(min_len))), True)

    def debug(self, message="", n=-1):
        if not self.__DEBUG_PLOT_DATAFRAME:
            return

        self.verify()

        if n > 0:
            sub_data_frame = self.data_frame.rx(
                robj.IntVector((range(n))), True)
        else:
            sub_data_frame = self.data_frame
        logger.debug(
            "data_frame: %s\n%s" % (message, str(sub_data_frame)))

    def verify(self):
        common.verifyVarType(common.isNotNone, self.__data_frame)

    def create(self, name_list, data_matrix, time_format=None):
        common.verifyVarType(common.isListOrTuple, name_list)
        common.verifyVarType(common.isListOrTuple, data_matrix)

        # todo:
        if common.isEmptyListOrTuple(data_matrix):
            raise ValueError("empty data")

        if len(name_list) != len(data_matrix):
            message = "not match length: field-name=%d, data-col=%d" % (
                len(name_list), len(data_matrix))
            raise ValueError(message)

        logger.debug(
            "create dataframe: names=(%s), #col=%d, #row=%d" % (
                ", ".join(name_list), len(data_matrix), len(data_matrix[0])))

        dict_name_datalist = {}

        for name, data_list in zip(name_list, data_matrix):
            is_datetime = name in (
                common.AN_SamplingStartTime, common.AN_SamplingEndTime)

            if is_datetime:
                value_type = RType.DATETIME
            else:
                value_type = RType.getRTypeCode(data_list)

            # convert data to R data format ---
            data_list = convertDataToR(data_list, value_type, time_format)
            dict_name_datalist[toDataFrameName(name)] = data_list

        self.__dict_name_datalist = dict(dict_name_datalist)
        self.__data_frame = robj.DataFrame(dict_name_datalist)

    def hasColumn(self, col_name):
        self.verify()

        common.verifyVarType(common.isNotEmptyString, col_name)

        if self.data_frame.rx2(toDataFrameName(col_name)) is rpy2.rinterface.RNULLType():
            logger.debug("data frame field not found: " + col_name)
            return False

        return True

    def verifyColumn(self, col_name):
        if not self.hasColumn(col_name):
            raise ValueError("'%s' column not found" % (col_name))

    def getColumnIndex(self, attr_name):
        i = 0
        for colname in self.data_frame.colnames:
            if colname == attr_name:
                break
            i += 1
        else:
            i = None

        return i

    def getMinMax(self, col_name):
        kargs = {"data": self.data_frame.rx2(col_name), "na.rm": True}

        min_data_y = robj.r["min"](**kargs)[0]
        max_data_y = robj.r["max"](**kargs)[0]

        return (min_data_y, max_data_y)

    def getUniqueList(self, col_name):
        """
        r_vector	= self.data_frame.rx2(col_name)
        fac_vector	= robj.FactorVector(r_vector)
        return tuple(fac_vector.levels)
        """
        return getUniqueList(self.data_frame.rx2(col_name))

    def getUniqueCount(self, col_name):
        #r_vector	= self.data_frame.rx2(col_name)
        # return len(getUniqueList(r_vector))
        return getUniqueCount(self.data_frame.rx2(col_name))

    def getTimeRange(self):
        import datetime

        sampling_time_field = toDataFrameName(common.AN_SamplingEndTime)

        if not self.hasColumn(sampling_time_field):
            return gtime.DateTimeRange(None, None)

        sampling_time_vector = self.data_frame.rx2(sampling_time_field)
        kargs = {"data": sampling_time_vector, "na.rm": True}
        min_timestamp = robj.r["min"](**kargs)[0]
        max_timestamp = robj.r["max"](**kargs)[0]
        min_datetime = datetime.datetime.fromtimestamp(min_timestamp)
        max_datetime = datetime.datetime.fromtimestamp(max_timestamp)

        return gtime.DateTimeRange(min_datetime, max_datetime)

    def melt(
            self, target_field_name_list,
            variable_name="variable", value_name="value"):
        """
        target_field_name_list: 集約対象の列名リスト
        variable_name: 集約後のKey列名
        value_name: 集約後のValue列名
        """

        self.debug()
        kargs = {
            "measure.vars"	:
            robj.StrVector(map(toDataFrameName, target_field_name_list)),
                "variable.name"	: toDataFrameName(variable_name),
                "value.name"	: toDataFrameName(value_name),
        }
        logger.debug("melt: " + common.dumpDict(kargs))

        kargs["data"] = self.data_frame

        self.__data_frame = robj.r["melt"](**kargs)

    def sort(self, sort_field, data_field):
        i = self.getColumnIndex(sort_field)
        if i is None:
            logger.warn("can not sord: " + sort_field)
            return

        # robj.r("""
        #	my_sort <- function(data_frame, order_key) {
        #		return(data_frame[order(data_frame$order_key),])
        #	}""")

        self.__data_frame[i] = robj.r["reorder"](**{
            "x"		: self.data_frame.rx2(sort_field),
            "X"		: self.data_frame.rx2(data_field),
            "FUN"	: "mean",
        })
        #data_frame[i]	= robj.r["rev"](data_frame[i])


def toDataFrameName(name):
    if common.isEmptyString(name):
        return None

    #name	= name.replace("%", "Percent_")
    name = re.sub("^CPU$|^cpu$", "cpu_id", name)
    name = name.replace("%", "")
    name = name.replace("/", "_per_")
    name = re.sub("[\s]+", "_", name)
    name = re.sub("[\+]", "_plus_", name)
    name = re.sub("[\-\(\)\:]", ".", name)

    return name


class RType:
    R_MAX_INT = 2 ** 31

    INT = 1
    FLOAT = 2
    TEXT = 3
    DATETIME = 4

    @classmethod
    def getRTypeCode(cls, data_list):
        type_code = cls.INT

        for data in data_list:
            if type_code == cls.INT:
                if common.isInteger(data):
                    if math.fabs(float(data)) >= cls.R_MAX_INT:
                        type_code = cls.FLOAT
                    continue

            if common.isFloat(data):
                type_code = cls.FLOAT
                continue

            return cls.TEXT

        return type_code


def convertDataToR(data_list, value_type, time_format):
    if value_type == RType.FLOAT:
        data_list = robj.FloatVector(data_list)
    elif value_type == RType.INT:
        data_list = robj.IntVector(data_list)
    elif value_type == RType.TEXT:
        # data_list	= robj.StrVector(toDataFrameNameList(data_list))
        data_list = robj.StrVector(data_list)
    elif value_type == RType.DATETIME:
        if common.isEmptyString(time_format):
            data_list = robj.FloatVector(data_list)
        else:
            data_list = [str(data) for data in data_list]
            if not gtime.isValidTimeFormat(data_list[0], time_format):
                logger.error(
                    "can not convert x axis values to R time format: " +
                    str(data_list))
                return None
            data_list = robj.r.strptime(data_list, time_format, tz="")
    else:
        logger.error("unknown type")

    return data_list


def toDatetimeToRobjTime2(org_datetime):
    if not gtime.isDatetime(org_datetime):
        return None

    time_format = "%Y%m%d %H%M%S"
    data = org_datetime.strftime(time_format)
    return robj.r.strptime(data, time_format, tz="")


def r_init():
    logger.debug("initialize R")

    robj.r("""
		library('ggplot2')
		library('scales')
		library('reshape2')	# for melt
		
		melt_wrapper <- function(data_frame, vars) {
			return(melt(data_frame, id.vars=vars))
		}""")

# class ---


class PlotFlag(common.BaseObject):

    def __init__(self):
        self.setAllFlag(True)

    def setAllFlag(self, plot_flag):
        self.is_plot_boxplot = plot_flag
        self.is_plot_linechart = plot_flag
        self.is_plot_scatter_diagram = plot_flag
        self.is_plot_stack_graph = plot_flag


class PlotParameterEx(common.BaseObject):
    __DEF_DPI = 150
    #__DEF_PlotFontFamily	= 'serif'
    __DEF_PlotFontFamily = 'MS Gothic'

    def getYAxisLabel(self):
        #data_list	= [self.y_axis_label, self.y_data_field]
        data_list = [self.y_axis_label]
        for data in data_list:
            if common.isNotEmptyString(data):
                return toDataFrameName(data)
        return ""

    def getXAxisLabel(self):
        #data_list	= [self.x_axis_label, self.x_data_field]
        data_list = [self.x_axis_label]
        for data in data_list:
            if common.isNotEmptyString(data):
                return toDataFrameName(data)
        return ""

    def getYDataField(self):
        data_list = [self.y_data_field, self.y_axis_label]
        for data in data_list:
            if common.isNotEmptyString(data):
                return toDataFrameName(data)
        return ""

    def getXDataField(self):
        data_list = [self.x_data_field, self.x_axis_label]
        for data in data_list:
            if common.isNotEmptyString(data):
                return toDataFrameName(data)
        return ""

    def __init__(self):
        self.root_path = ""

        self.table_name = ""
        self.y_data_field = None
        self.x_data_field = None
        self.group_field_name = None
        self.group_field_factor = None
        self.unit_name = None

        # ----
        self.output_dir = None
        self.file_name_sufix = None
        self.image_file_format = EXT_PNG
        self.image_width_pixel = None		# None: auto modified
        self.image_height_pixel = None		# None: auto modified

        self.font_family = self.__DEF_PlotFontFamily

        self.title = ""
        self.title_prefix = ""
        self.title_suffix = ""

        self.x_axis_label = ""
        self.y_axis_label = ""
        self.y_breaks = None
        self.y_minor_breaks = None

        self.y_log = None		# None, log2, log10
        # self.log	= None	# "x", "y", "xy"
        self.y_max = None
        self.sum_y_max = None

        # for boxplot
        self.horizontal = False
        self.notch = False

        # line chart parameter ---
        # todo: create new class
        self.time_format = None		# for x axis

        self.is_plot_point = True
        self.is_plot_line = True

        # stack graph parameter ---
        # todo: create new class
        self.stack_style = PlotType.Stack.AREA
        self.order = None
        self.facet_grid = None

        # rect parameter ---
        # todo: create new class
        self.xmin_vector = None
        self.xmax_vector = None
        self.ymin_vector = None
        self.ymax_vector = None

        # tmp ---
        self.dpi = self.__DEF_DPI
        self.col = None
        self.xlim = None		# tuple
        self.ylim = None		# tuple

        self.shape = None
        self.fill = None
        self.linetype = None
        self.group = None
        self.size = None

        self.legend_title = None
        self.scale_labels = None
        self.num_vertical_facet = 1
        # self.num_horizontal_facet

        self.unique_item_count = None

    def isValidUnit(self):
        if common.isEmptyString(self.unit_name):
            return False

        return self.unit_name != sysstat.Unit.DIMENSIONLESS

    def validation(self):
        common.verifyVarType(common.isNotEmptyString, self.getXDataField())
        common.verifyVarType(common.isNotEmptyString, self.getYDataField())

    def isValidParameter(self):
        if common.isEmptyString(self.table_name):
            logger.debug("empty table name")
            return False

        if common.isEmptyString(self.group_field_name):
            logger.debug("empty group field name")
            return False

        if common.isEmptyString(self.group_field_factor):
            logger.debug("empty group field factor")
            return False

        return True

    def getTitleElementList(self):
        title_list = []
        if common.isNotEmptyString(self.title_prefix):
            title_list.append(self.title_prefix)
        title_list.append(self.table_name)
        # title_list.append(self.group_field_name)
        if common.isNotEmptyString(self.title_suffix):
            title_list.append(self.title_suffix)

        return title_list

    def getPlotFileBaseName(self):
        fname_item_list = []

        if common.isNotEmptyString(self.title):
            fname_item_list.append(self.title)

        if common.isNotEmptyString(self.file_name_sufix):
            fname_item_list.append(self.file_name_sufix)

        if common.isEmptyListOrTuple(fname_item_list):
            return ""

        base_fname = "_".join(fname_item_list)
        base_fname = gfile.adjustFileName(base_fname)

        return base_fname

    def makePlotFileName(self, plot_type):
        if common.isEmptyString(self.output_dir):
            raise ValueError("null output directory path")

        base_fname = self.getPlotFileBaseName() + "_" + plot_type

        if os.path.exists(self.output_dir):
            if not os.path.isdir(self.output_dir):
                logger.error(
                    "different file already exists: " + self.output_dir)
        else:
            gfile.FileManager.makeDirectory(self.output_dir)

        file_id = 0
        while True:
            if file_id == 0:
                file_name = "%s.%s" % (base_fname, self.image_file_format)
            else:
                file_name = "%s_%02d.%s" % (
                    base_fname, file_id, self.image_file_format)

            full_path = os.path.sep.join([self.output_dir, file_name])
            if os.path.exists(full_path):
                file_id += 1
                continue
            break

        return full_path

    def getScale(self, scale_args=None):
        if scale_args is None:
            scale_args = {}

        # if not common.isEmptyListOrTuple(self.ylim):
        #	scale_args["limits"]	= robj.FloatVector(self.ylim)

        if not common.isEmptyListOrTuple(self.y_breaks):
            scale_args["breaks"] = robj.FloatVector(self.y_breaks)
        if not common.isEmptyListOrTuple(self.scale_labels):
            scale_args["labels"] = robj.StrVector(self.scale_labels)
        if common.isNotEmptyString(self.legend_title):
            scale_args["name"] = self.legend_title

        common.debugDict(scale_args, locals())

        return ggplot2.scale_y_continuous(**scale_args)

    def getCoordTrans(self):
        logger.debug("get coord_trans: " + str(self.y_log))

        dict_coord_args = {}

        if common.isNotEmptyString(self.y_log):
            if self.y_log not in ("log2", "log10"):
                logger.error("unknown log scale: " + self.y_log)
                return None

            dict_coord_args["y"] = self.y_log

        if not common.isEmptyListOrTuple(self.ylim):
            dict_coord_args["limy"] = robj.FloatVector(self.ylim)

        common.debugDict(dict_coord_args, locals())

        if len(dict_coord_args) == 0:
            return None

        return ggplot2.coord_trans(**dict_coord_args)

    def getAesStringArgs(self):
        aes_string_args = {}

        aes_string_args["x"] = self.getXDataField()
        aes_string_args["y"] = self.getYDataField()

        if common.isNotEmptyString(self.col):
            aes_string_args["col"] = toDataFrameName(self.col)
        if common.isNotEmptyString(self.fill):
            aes_string_args["fill"] = toDataFrameName(self.fill)
        if common.isNotEmptyString(self.shape):
            aes_string_args["shape"] = toDataFrameName(self.shape)
        if common.isNotEmptyString(self.linetype):
            aes_string_args["linetype"] = toDataFrameName(self.linetype)
        if common.isNotEmptyString(self.group):
            aes_string_args["group"] = toDataFrameName(self.group)
        if common.isNotEmptyString(self.size):
            aes_string_args["size"] = "factor(%s)" % (
                toDataFrameName(self.size))

        """
        if common.isFloat(self.ymin):
        	aes_string_args["ymin"]	= str(self.ymin)
        if common.isFloat(self.ymax):
        	aes_string_args["ymax"]	= str(self.ymax)
        	"""

        common.debugDict(aes_string_args, locals())

        return aes_string_args

    def getNumColumnLegend(self, width_pixel, data_frame):
        if common.isEmptyString(self.group):
            raise ValueError("getNumColumnLegend: empty group")

        legend_len_list = []
        for legend_item in data_frame.getUniqueList(toDataFrameName(self.group)):
            legend_len_list.append(len(legend_item))

        import golib.statistics as stats

        dict_stat = stats.Calculator.getStatistics(legend_len_list)
        row_size_pixel = math.ceil(dict_stat.get(stats.MEAN) * 16)

        return int(math.ceil((width_pixel - 200) / row_size_pixel))

    def getImageHeightPixel(self, width_pixel, data_frame):
        if common.isFloat(self.image_height_pixel):
            return self.image_height_pixel

        message_list = []

        if common.isNotEmptyString(self.facet_grid):
            height_pixel = self.num_vertical_facet * 180 + 360
            message_list.append("y-facet=%d" % (self.num_vertical_facet))
        else:
            if common.isEmptyListOrTuple(self.y_breaks):
                height_pixel = 820
            else:
                height_pixel = 560 + 20 * len(self.y_breaks)
                message_list.append("y-count=%d" % (len(self.y_breaks)))

        if common.isInteger(self.unique_item_count):
            legend_row = math.ceil(
                self.unique_item_count / self.getNumColumnLegend(width_pixel, data_frame))
            height_pixel += legend_row * 16

        message_list.append("height-pixel=%d" % (height_pixel))
        logger.debug(", ".join(message_list))

        return height_pixel

    def getImageHeightInch(self, data_frame):
        width_pixel = self.getImageWidthPixel(data_frame)
        height_pixel = self.getImageHeightPixel(width_pixel, data_frame)

        return height_pixel / float(self.dpi)

    def getImageWidthPixel(self, data_frame):
        x_axis_label = self.getXDataField()

        if common.isNotEmptyString(x_axis_label):
            x_count = data_frame.getUniqueCount(x_axis_label)
        else:
            logger.debug("getImageWidthPixel: empty x axis label")
            x_count = 0

        if common.isFloat(self.image_width_pixel):
            width_pixel = self.image_width_pixel
        else:
            width_pixel = 480 + 16 * x_count
            width_pixel = max(width_pixel, MIN_ImageWidthPixel)
            width_pixel = min(width_pixel, MAX_ImageWidthPixel)

        logger.debug(
            "width-pixel=%d, x-count=%d" % (width_pixel, x_count))

        return width_pixel

    def getImageWidthInch(self, data_frame):
        return self.getImageWidthPixel(data_frame) / float(self.dpi)

    def getPartiallyColoredVector(self):
        vector_size = self.unique_item_count
        max_colored = 24
        #color_count	= math.ceil(vector_size * 0.3)
        color_count = min([max_colored, vector_size])

        return getColorVector(color_count, vector_size)

    def getGuideLegend(self, width_pixel, data_frame):
        legend_args = {}
        legend_args["ncol"] = self.getNumColumnLegend(width_pixel, data_frame)

        title = None
        if common.isNotEmptyString(self.legend_title):
            title = self.legend_title
            legend_args["title"] = title

        common.debugDict(legend_args, locals())

        return ggplot2.guide_legend(**legend_args)


class ScaleBreaks:
    __KILO = 1024.0
    __BYTE_UNIT_LIST = (
        sysstat.Unit.BYTE,
        sysstat.Unit.KB,
        sysstat.Unit.MB,

        sysstat.Unit.BYTE_PER_SEC,
        sysstat.Unit.KB_PER_SEC,
        sysstat.Unit.MB_PER_SEC,

        sysstat.Unit.KB_PER_PACKET,
        sysstat.Unit.KB_PER_TRANS,
    )

    @classmethod
    def __getEvenBreaks(
            cls, min_value, max_value, div=10, is_include_boundary=True):
        step = (max_value - min_value) / float(div)
        breaks = []
        if is_include_boundary:
            seq = range(div + 1)
        else:
            seq = range(1, div)

        for i in seq:
            breaks.append(min_value + step * i)

        logger.debug("even breaks: min=%f, max=%f, div=%d, breaks=%s" % (
            min_value, max_value, div, str(breaks)))

        return breaks

    @classmethod
    def __getMinorEvenBreaks(cls, breaks, div=10):
        minor_breaks = set()
        for i in range(len(breaks) - 1):
            range_min = breaks[i]
            range_max = breaks[i + 1]

            for minor_break in cls.__getEvenBreaks(range_min, range_max, div, is_include_boundary=False):
                minor_breaks.add(minor_break)

        logger.debug(
            "minor breaks of %s: %s" % (str(breaks), str(minor_breaks)))

        return list(minor_breaks)

    @classmethod
    def __getSign(cls, value):
        sign_table = {
            True: 1,
            False: -1,
        }

        return sign_table[value >= 0]

    @classmethod
    def __getLimitMinPower(cls, min_value, base):
        power = math.log(math.fabs(min_value), base)

        func_table = {
            True: {
                True: math.floor,
                False: math.ceil,
            },
            False: {
                True: math.ceil,
                False: math.floor,
            }
        }

        abs_min_power = func_table[min_value >= 0][
            power >= 0](math.fabs(power))

        # if math.fabs(abs_min_power - abs_power) < 0.1:
        #	abs_min_power	-= sign_power

        return int(abs_min_power * cls.__getSign(power))

        """
        try:
        	power		= math.log(math.fabs(min_value), base)
        	abs_power	= math.fabs(power)
        	
        	if min_value >= 0:
        		sign_value	= 1
        	else:
        		sign_value	= -1
        	
        	if power >= 0:
        		sign_power	= 1
        		abs_min_power	= math.floor(abs_power)
        	else:
        		sign_power	= -1
        		abs_min_power	= math.ceil(abs_power)
        
        	if math.fabs(abs_min_power - abs_power) < 0.1:
        		abs_min_power	-= sign_power
        	
        	print abs_min_power, sign_power, sign_value
        	return int(abs_min_power * sign_power * sign_value)
        except Exception, e:
        	logger.exception(e, "value=%s" % (str(min_value)))
        	return None
        	"""

    @classmethod
    def __getLimitMaxPower(cls, max_value, base):
        power = math.log(math.fabs(max_value), base)

        func_table = {
            True: {
                True: math.ceil,
                False: math.floor,
            },
            False: {
                True: math.floor,
                False: math.ceil,
            }
        }

        abs_max_power = func_table[max_value >= 0][
            power >= 0](math.fabs(power))

        # if math.fabs(abs_max_power - abs_power) < 0.1:
        #	abs_max_power	+= sign

        return int(abs_max_power * cls.__getSign(power))

        """
		try:
			power		= math.log(max_value, base)
			abs_power	= math.fabs(power)
			if power > 0:
				sign	= 1.0
				abs_max_power	= math.ceil(abs_power)
			else:
				sign	= -1.0
				abs_max_power	= math.floor(abs_power)
		
			if math.fabs(abs_max_power - abs_power) < 0.1:
				abs_max_power	+= sign
			
			return int(abs_max_power * sign)
		except Exception, e:
			logger.exception(e, "value=%s" % (str(max_value)))
			return None
			"""

    @classmethod
    def __getLog10MinorBreaksDiv(cls, num_breaks, is_log=False):
        if num_breaks >= 100:
            return 2
        elif num_breaks >= 40:
            if is_log:
                return 4
            return 5

        if is_log:
            return 9
        return 10

    @classmethod
    def __getLog2MinorBreaksDiv(cls, num_breaks):
        if num_breaks >= 48:
            return 2
        elif num_breaks >= 12:
            return 4
        else:
            return 8

    @classmethod
    def __getLogBreaks(cls, min_value, max_value, base):
        if min_value > max_value:
            raise ValueError("min > max")

        # if min_value == 0:
        #	min_value	= -1

        limit_max_power = cls.__getLimitMaxPower(max_value, base)

        if min_value != 0:
            limit_min_power = cls.__getLimitMinPower(min_value, base)
        else:
            limit_min_power = limit_max_power - 4
            limit_min_power = min(limit_min_power, 0)

        breaks = []

        if max_value < 0:
            for power in range(limit_min_power, limit_max_power - 1, -1):
                breaks.append(math.pow(base, power) * cls.__getSign(min_value))
        elif min_value < 0:
            min_power = 0

            for power in range(abs(limit_min_power), min_power - 1, -1):
                breaks.append(math.pow(base, power) * cls.__getSign(min_value))

            for power in range(min_power, limit_max_power + 1):
                breaks.append(math.pow(base, power) * cls.__getSign(max_value))
        else:
            for power in range(limit_min_power, limit_max_power + 1):
                breaks.append(math.pow(base, power) * cls.__getSign(min_value))

        """
		max_power	= cls.__getLimitMaxPower(max_value, base)
		
		if min_value > 0:
			min_power	= cls.__getLimitMinPower(min_value, base)
		else:
			if base >= 10:
				min_power	= max_power - 4
			else:
				min_power	= max_power - 8
			min_power	= min(min_power, 0)
			
		if min_power == max_power:
			min_power	-= 1
			max_power	+= 1
		
		breaks	= []
		for power in range(min_power, max_power + 1):
			breaks.append(math.pow(base, power))
		"""

        logger.debug(
            "log %d breaks: min=%f, max=%f, breaks=%s" % (
                base, min_value, max_value, str(breaks)))

        return breaks

    @classmethod
    def __getScalerByteLabels(cls, byte_breaks):
        return [
            common.convertBytesToHumanReadable(int(byte)) for byte in byte_breaks]

    @classmethod
    def __getBytePerSecLabels(cls, kb_breaks):
        return [kb_label + "/s" for kb_label in cls.__getScalerByteLabels(kb_breaks)]

    @classmethod
    def __getBytePerPacketLabels(cls, kb_breaks):
        return [
            kb_label + "/packet" for kb_label in cls.__getScalerByteLabels(kb_breaks)]

    @classmethod
    def __getPercentLabels(cls, percent_breaks):
        return [str(item) + "%" for item in percent_breaks]

    @classmethod
    def __convertBreaksToLabelsScaler(cls, unit_name, breaks):
        if unit_name == sysstat.Unit.BYTE:
            byte_breaks = breaks
        elif unit_name in [sysstat.Unit.KB, sysstat.Unit.KB_PER_PACKET, sysstat.Unit.KB_PER_TRANS]:
            byte_breaks = [kb * cls.__KILO for kb in breaks]
        elif unit_name == sysstat.Unit.MB:
            byte_breaks = [mb * (cls.__KILO ** 2) for mb in breaks]
        else:
            raise ValueError("unknown unit name: " + unit_name)

        return cls.__getScalerByteLabels(byte_breaks)

    @classmethod
    def __convertBreaksToLabelsBytePerSec(cls, unit_name, breaks):
        if unit_name == sysstat.Unit.BYTE_PER_SEC:
            byte_breaks = breaks
        elif unit_name == sysstat.Unit.KB_PER_SEC:
            byte_breaks = [kb * cls.__KILO for kb in breaks]
        elif unit_name == sysstat.Unit.MB_PER_SEC:
            byte_breaks = [mb * (cls.__KILO ** 2) for mb in breaks]
        else:
            raise ValueError("unknown unit name: " + unit_name)

        return cls.__getBytePerSecLabels(byte_breaks)

    @classmethod
    def __convertBreaksToLabelsBytePerPacket(cls, unit_name, breaks):
        byte_breaks = [kb * cls.__KILO for kb in breaks]

        return cls.__getBytePerPacketLabels(byte_breaks)

    @classmethod
    def __convertBreaksToLabels(cls, unit_name, breaks):
        byte_unit_list = [
            sysstat.Unit.BYTE,
            sysstat.Unit.KB,
            sysstat.Unit.MB,

            sysstat.Unit.KB_PER_TRANS,
            sysstat.Unit.KB_PER_PACKET,
        ]

        if unit_name in byte_unit_list:
            scale_labels = cls.__convertBreaksToLabelsScaler(unit_name, breaks)
        elif unit_name in [sysstat.Unit.BYTE_PER_SEC, sysstat.Unit.KB_PER_SEC, sysstat.Unit.MB_PER_SEC]:
            scale_labels = cls.__convertBreaksToLabelsBytePerSec(
                unit_name, breaks)
        # elif unit_name in [sysstat.Unit.KB_PER_PACKET]:
        #	scale_labels	= cls.__convertBreaksToLabelsBytePerPacket(unit_name, breaks)
        else:
            raise ValueError("unknown unit name: " + unit_name)

        return scale_labels

    @classmethod
    def __adjustLimitMaxY(cls, y_max, input_y_limit, div_count):
        """
        input_y_limitを上限
        0～input_y_limit
        """

        common.debugFunc(cls.__adjustLimitMaxY, locals())

        divider = common.safeDivision(input_y_limit, div_count)
        return_limit = math.ceil(y_max / divider) * divider
        major_div = int(return_limit / divider)
        return major_div, return_limit

        """
		major_div			= 1
		step				= common.safeDivision(input_y_limit, div_count)
		adjust_y_limit_max	= step
		while adjust_y_limit_max <= y_max:
			adjust_y_limit_max	+= step
			major_div	+= 1
			
		logger.debug("adjust y limit: div=%d, output-y-limit=%d" % (
			major_div, adjust_y_limit_max))
		
		return major_div, adjust_y_limit_max
		"""

    @classmethod
    def getDistParameterEx(cls, unit_name, min_data_y, max_data_y, num_y_facet):
        logger.debug(
            "getDistParameter: unit=%s, min-y=%s, max-y=%s, facert-y=%s" % (
                unit_name, str(min_data_y), str(max_data_y), str(num_y_facet)))

        if unit_name in cls.__BYTE_UNIT_LIST:
            breaks = cls.__getLogBreaks(0, max_data_y, base=4)
            # print "!!", breaks
            minor_div = cls.__getLog2MinorBreaksDiv(len(breaks) * num_y_facet)
            y_minor_breaks = cls.__getMinorEvenBreaks(breaks, minor_div)
            y_log = "log2"

            scale_labels = cls.__convertBreaksToLabels(unit_name, breaks)
        elif unit_name == sysstat.Unit.PERCENT:
            y_limit_max = max(cls.__getLogBreaks(0, max_data_y, base=10))
            y_limit_max = min(y_limit_max, 100)
            breaks = cls.__getLogBreaks(0, y_limit_max, base=10)
            minor_div = cls.__getLog10MinorBreaksDiv(
                len(breaks) * num_y_facet, is_log=True)
            y_minor_breaks = cls.__getMinorEvenBreaks(breaks, minor_div)
            scale_labels = cls.__getPercentLabels(breaks)
            y_log = "log10"
        else:
            breaks = cls.__getLogBreaks(min_data_y, max_data_y, base=10)
            minor_div = cls.__getLog10MinorBreaksDiv(
                len(breaks) * num_y_facet, is_log=True)
            y_minor_breaks = cls.__getMinorEvenBreaks(breaks, minor_div)
            scale_labels = None
            y_log = "log10"

        # print "##", unit_name, min_data_y, max_data_y, breaks
        ylim = (min(breaks), max(breaks))

        logger.debug(
            "breaks=%s, labels=%s" % (str(breaks), str(scale_labels)))

        return breaks, y_minor_breaks, ylim, scale_labels, y_log

    @classmethod
    def getStackBarParameter(cls, unit_name, y_min, y_max, num_y_facet):
        common.verifyVarType(common.isFloat, y_max, locals())
        # if not common.isFloat(y_max):
        #	raise ValueError("float is required: y_max")

        if unit_name in cls.__BYTE_UNIT_LIST:
            y_limit_min = 0
            y_limit_max = max(cls.__getLogBreaks(0, y_max, base=2))
            dummy, y_limit_max = cls.__adjustLimitMaxY(
                y_max, y_limit_max, div_count=8)
            major_div = 8

            y_breaks = cls.__getEvenBreaks(
                y_limit_min, y_limit_max, div=major_div)
            minor_div = cls.__getLog2MinorBreaksDiv(
                len(y_breaks) * num_y_facet)
            #y_minor_breaks	= cls.__getMinorEvenBreaks(y_breaks, minor_div)

            scale_labels = cls.__convertBreaksToLabels(unit_name, y_breaks)
        else:
            base = 10
            log_breaks = cls.__getLogBreaks(y_min, y_max, base)
            max_div, y_limit_max = cls.__adjustLimitMaxY(
                y_max, max(log_breaks), base)

            min_div = 0
            y_limit_min = 0

            if unit_name == sysstat.Unit.PERCENT:
                y_limit_max = min(y_limit_max, 100)
                if y_limit_max == 100:
                    if num_y_facet >= 8:
                        max_div = 2
                    elif num_y_facet >= 4:
                        max_div = 5
                    else:
                        max_div = 10
            else:
                if y_min != 0:
                    min_div, y_limit_min = cls.__adjustLimitMaxY(
                        y_min, min(log_breaks), base)

            y_breaks = cls.__getEvenBreaks(
                y_limit_min, y_limit_max, div=min_div + max_div)

            if unit_name == sysstat.Unit.PERCENT:
                scale_labels = cls.__getPercentLabels(y_breaks)
            else:
                scale_labels = None

            minor_div = cls.__getLog10MinorBreaksDiv(
                len(y_breaks) * num_y_facet)
            #y_minor_breaks	= cls.__getMinorEvenBreaks(y_breaks, minor_div)

        """
		elif unit_name == sysstat.Unit.PERCENT:
			base	= 10
			y_limit_max	= max(cls.__getLogBreaks(0, y_max, base))
			major_div, y_limit_max	= cls.__adjustLimitMaxY(y_max, y_limit_max, base)
			y_limit_max		= min(y_limit_max, 100)
			if y_limit_max == 100:
				if num_y_facet >= 8:
					major_div	= 2
				elif num_y_facet >= 4:
					major_div	= 5
				else:
					major_div	= 10
			
			y_breaks		= cls.__getEvenBreaks(y_limit_min, y_limit_max, div=major_div)
			minor_div		= cls.__getLog10MinorBreaksDiv(len(y_breaks) * num_y_facet)
			#y_minor_breaks	= cls.__getMinorEvenBreaks(y_breaks, minor_div)
			
			scale_labels	= cls.__getPercentLabels(y_breaks)
			"""

        y_minor_breaks = cls.__getMinorEvenBreaks(y_breaks, minor_div)
        ylim = (min(y_breaks), max(y_breaks))

        return y_breaks, y_minor_breaks, ylim, scale_labels

# ggplot2 ---

if LOAD_GGPLOT2:
    def getPlotter(plot_type, plot_param, dry_run):
        if plot_type == PlotType.SCATTER_DIAGRAM:
            return GGPlotScatterPlotter(plot_param, dry_run)
        if plot_type == PlotType.LINE_CHART:
            return GGPlotScatterPlotter(plot_param, dry_run)
        if plot_type == PlotType.BOXPLOT:
            return GGPlotBoxplotPlotter(plot_param, dry_run)
        if plot_type == PlotType.STACK_GRAPH:
            return GGPlotStackBarPlotter(plot_param, dry_run)
        if plot_type == PlotType.STACK_BAR_GRAPH:
            return GGPlotStackBarPlotter(plot_param, dry_run)
        if plot_type == PlotType.RECT:
            return GGPlotRectPlotter(plot_param, dry_run)

        if plot_type == PlotType.BAR_GRAPH:
            raise NotImplementedError(plot_type)

        if plot_type == PlotType.HISTGRAM:
            raise NotImplementedError(plot_type)

        raise NotImplementedError()

    # class ---

    class GGPlotBasePlotter(common.BaseObject):
        _DEFAULT_THEME = {
            #'legend.title':			ggplot2.element_blank(),
            'legend.title'			: ggplot2.element_text(size=8),
            'legend.text'			: ggplot2.element_text(size=6),
            "legend.position"		: "bottom",
            "legend.key.size"		: grid.unit(0.35, "cm"),

            "panel.border"			:
            ggplot2.element_rect(fill="NA", linetype="solid", colour="gray"),
                "panel.grid.minor.y"	: ggplot2.element_blank(),
                "panel.margin"			: grid.unit(0.4, "cm"),
                #"panel.grid.minor.x"	: ggplot2.element_blank(),
                #"axis.text.y":			ggplot2.element_text(size=6),
                #"axis.text.y"	: robj.r["rel"](0.8),
        }

        @property
        def plot_param(self):
            return self.__plot_param

        @property
        def dry_run(self):
            return self.__dry_run

        def __init__(self, plot_param, dry_run):
            self.__plot_param = plot_param
            self.__dry_run = dry_run

            # rpy2.rinterface.set_writeconsole(None)
            # r_init()

        def save(self, pp, data_frame):
            save_path = self.getPlotFileName()

            qqsave_args = {
                "plot"		: pp,
                "file"		: save_path,
                "dpi"		: self.plot_param.dpi,
                "limitsize"	: False,
                "width"		: self.plot_param.getImageWidthInch(data_frame),
                "height"	: self.plot_param.getImageHeightInch(data_frame),
            }
            common.debugDict(qqsave_args, locals())

            if self.dry_run:
                return save_path

            try:
                robj.r["ggsave"](**qqsave_args)
            except rpy2.rinterface.RRuntimeError, e:
                logger.exception(e)
                return None

            return save_path

        def _ggplotBasic(
                self, data_frame, aes_string_args, theme_args, guide_legend):

            common.debugDict(aes_string_args, locals())
            common.debugDict(theme_args, locals())

            pp = ggplot2.ggplot(data_frame.data_frame)
            pp = pp + ggplot2.aes_string(**aes_string_args)
            pp = pp + ggplot2.theme(**theme_args)
            pp = pp + ggplot2.guides(col=guide_legend)

            scale = self.plot_param.getScale()
            if scale:
                pp = pp + scale

            minor_breaks = self.__plotMinorBreaks()
            if minor_breaks:
                pp = pp + minor_breaks

            #"""
            coord_trans = self.plot_param.getCoordTrans()
            if coord_trans:
                pp = pp + coord_trans
            #"""

            if common.isNotEmptyString(self.plot_param.facet_grid):
                pp = pp + ggplot2.facet_grid(self.plot_param.facet_grid)
            if common.isNotEmptyString(self.plot_param.title):
                pp = pp + ggplot2.ggtitle(self.plot_param.title)

            pp = pp + ggplot2.labs(
                x=self.plot_param.getXAxisLabel(),
                y=self.plot_param.getYAxisLabel())

            return pp

        def __plotMinorBreaks(self):
            if common.isEmptyListOrTuple(self.plot_param.y_minor_breaks):
                logger.debug("empty minor breaks: y axis")
                return None

            line_color = "#f4f4f4"

            return ggplot2.geom_hline(
                yintercept=robj.FloatVector(self.plot_param.y_minor_breaks),
                size=0.2,
                colour=line_color)

        def getPlotFileName(self):
            return self.plot_param.makePlotFileName(self.getPlotType())

        def showPlotInfo(self, data_frame):
            FORMAT_PLOT_BEGIN = "plot %s: %s"

            logger.info(FORMAT_PLOT_BEGIN % (
                self.getPlotType(), self.getPlotFileName()))

            self.plot_param.debug()
            data_frame.debug(n=32)

        # abstract function ---

        def getPlotType(self):
            return ""

        def getWidthPixel(self, data_frame):
            return 0

        def plot(self, data_frame):
            pass

    class GGPlotTimelinePlotter(GGPlotBasePlotter):

        def __init__(self, plot_param, dry_run):
            super(GGPlotTimelinePlotter, self).__init__(plot_param, dry_run)

        def plotTimeRangeVerticalLine(self, pp, data_frame):
            common.verifyVarType(isRDataFrame, data_frame)

            datetime_range = data_frame.getTimeRange()
            vline_color = "black"
            vline_size = 0.5

            if gtime.isDatetime(datetime_range.start_datetime):
                min_rtime = toDatetimeToRobjTime2(
                    datetime_range.start_datetime)

                aes_vline = {
                    "xintercept"	: robj.r["as.numeric"](min_rtime),
                    #"xintercept"	: robj.IntVector([robj.r["as.numeric"](min_rtime), robj.r["as.numeric"](min_rtime)]),
                    "color"			: vline_color,
                    "size"			: vline_size,
                }
                common.debugDict(aes_vline, locals())
                pp = pp + ggplot2.geom_vline(**aes_vline)

            if gtime.isDatetime(datetime_range.end_datetime):
                max_rtime = toDatetimeToRobjTime2(datetime_range.end_datetime)

                aes_vline = {
                    "xintercept"	: robj.r["as.numeric"](max_rtime),
                    "color"			: vline_color,
                    "size"			: vline_size,
                }
                common.debugDict(aes_vline, locals())
                pp = pp + ggplot2.geom_vline(**aes_vline)

            return pp

    class GGPlotBoxplotPlotter(GGPlotBasePlotter):

        def __init__(self, plot_param, dry_run):
            super(GGPlotBoxplotPlotter, self).__init__(plot_param, dry_run)

        def getPlotType(self):
            return PlotType.BOXPLOT

        def __getBoxWidth(self, unique_item_count):
            """
            box_width	= 40
            if unique_item_count >= 300:
                    box_width	= 10
            elif unique_item_count >= 200:
                    box_width	= 12
            elif unique_item_count >= 100:
                    box_width	= 14
            elif unique_item_count >= 90:
                    box_width	= 16
            elif unique_item_count >= 80:
                    box_width	= 18
            elif unique_item_count >= 70:
                    box_width	= 20
            elif unique_item_count >= 60:
                    box_width	= 22
            elif unique_item_count >= 50:
                    box_width	= 24
            elif unique_item_count >= 40:
                    box_width	= 26
            elif unique_item_count >= 30:
                    box_width	= 28
            elif unique_item_count >= 20:
                    box_width	= 30
            elif unique_item_count >= 10:
                    box_width	= 32
            """

            min_width = 10
            max_width = 40

            width = max_width - \
                int(math.ceil(math.log(unique_item_count, 1.5) * 2.0))
            width = max(min_width, width)
            width = min(max_width, width)

            return width

        def __getBoxPlotWidth(self):
            item_count = self.plot_param.unique_item_count

            if not common.isInteger(item_count):
                return MIN_ImageWidthPixel

            base_width_pixel = 160
            width_pixel = item_count * \
                self.__getBoxWidth(item_count) + base_width_pixel
            width_pixel = max(width_pixel, MIN_ImageWidthPixel)

            return width_pixel

        def getWidthPixel(self, data_frame):
            return self.__getBoxPlotWidth()

        def plot(self, data_frame):
            """
            datasets = importr('datasets')
            mtcars = datasets.__rdata__.fetch('mtcars')['mtcars']
            print mtcars
            """

            common.verifyVarType(isRDataFrame, data_frame)
            self.plot_param.validation()
            self.showPlotInfo(data_frame)

            x_data_field = self.plot_param.getXDataField()

            if not data_frame.hasColumn(x_data_field):
                return

            data_frame.sort(
                sort_field=x_data_field,
                data_field=self.plot_param.getYDataField())
            aes_string_args = self.plot_param.getAesStringArgs()

            theme_args = dict(self._DEFAULT_THEME)
            theme_args['axis.text.x'] = ggplot2.element_text(angle=90, size=9)

            geom_boxplot_args = {
                "outlier.size"	: 1,
                "outlier.shape"	: 1,
                "alpha"			: 0.6,
                "notch"			: self.plot_param.notch,
                # "notchwidth"	: 0.2,
            }

            common.debugDict(geom_boxplot_args, locals())

            guide_legend = self.plot_param.getGuideLegend(
                self.getWidthPixel(data_frame), data_frame)

            # print common.dumpDict(theme_args)

            pp = self._ggplotBasic(
                data_frame, aes_string_args, theme_args, guide_legend)
            pp = pp + ggplot2.geom_boxplot(**geom_boxplot_args)
            if common.isNotEmptyString(self.plot_param.facet_grid):
                pp = pp + ggplot2.facet_grid(self.plot_param.facet_grid)

            pp = pp + ggplot2.stat_summary(**{
                "fun.y": "mean",
                "geom": "point",
                "shape": 23,
                "fill": "white",
            })

            """
			pp	= pp + ggplot2.scale_fill_gradientn(
							colours=robj.rainbow(30),
							guide=ggplot2.guide_legend(ncol=self.(width_pixel)))
			pp	= pp + ggplot2.scale_fill_brewer(
							type="qual",
							pallete=3,
							guide=ggplot2.guide_legend(ncol=self._getNumColumnLegend(width_pixel)))
							"""
            #pp	= pp + ggplot2.annotation_logticks(sides="lr")

            if common.isInteger(self.plot_param.unique_item_count):
                #color_count	= math.ceil(self.unique_item_count * 0.2)
                vector_size = self.plot_param.unique_item_count
                pp = pp + ggplot2.scale_fill_manual(
                    values=getGrayColorVector(vector_size, vector_size), guide=guide_legend)
                pp = pp + ggplot2.scale_colour_manual(
                    values=getColorVector(vector_size, vector_size), guide=guide_legend)

            """
			if not common.isEmptyListOrTuple(self.ylim):
				pp	= pp + ggplot2.coord_cartesian(ylim=robj.FloatVector(self.ylim))
				"""

            return self.save(pp, data_frame)

    class GGPlotScatterPlotter(GGPlotTimelinePlotter):

        def __init__(self, plot_param, dry_run):
            super(GGPlotScatterPlotter, self).__init__(plot_param, dry_run)

        def getPlotType(self):
            if self.plot_param.is_plot_line:
                return PlotType.LINE_CHART
            else:
                return PlotType.SCATTER_DIAGRAM

        def getWidthPixel(self, data_frame):
            return self.plot_param.getImageWidthPixel(data_frame)

        def plot(self, data_frame):
            # print [column.rclass[0] for column in data_frame]
            common.verifyVarType(isRDataFrame, data_frame)

            if all([not self.plot_param.is_plot_point, not self.plot_param.is_plot_line]):
                logger.error("plot type not defined")
                return

            self.showPlotInfo(data_frame)

            aes_string_args = self.plot_param.getAesStringArgs()
            theme_args = dict(self._DEFAULT_THEME)
            # if common.isNotEmptyString(self.legend_title):
            #	theme_args["legend.title"]	= self.legend_title
            theme_args["panel.grid.minor.y"] = ggplot2.element_blank()

            geom_point_args = {"alpha": 0.6}
            geom_line_args = {"alpha": 0.3}
            if aes_string_args.get("size") is None:
                geom_point_args["size"] = 1.5
                geom_line_args["size"] = 0.4
            common.debugDict(geom_point_args, locals())
            common.debugDict(geom_line_args, locals())

            guide_legend = self.plot_param.getGuideLegend(
                self.getWidthPixel(data_frame), data_frame)

            #datetime_list	= robj.r.strptime([min_datetime, max_datetime], time_format, tz="")

            pp = self._ggplotBasic(
                data_frame, aes_string_args, theme_args, guide_legend)
            pp = pp + ggplot2.scale_x_datetime()

            if self.plot_param.is_plot_point:
                pp = pp + ggplot2.geom_point(**geom_point_args)
            if self.plot_param.is_plot_line:
                pp = pp + ggplot2.geom_line(**geom_line_args)
            """
			else:
				pp	= pp + ggplot2.geom_smooth(
					method = 'lm', fullrange=True, fill="gray")
			"""

            # R 2.15ではエラーとなる
            #pp	= self.plotTimeRangeVerticalLine(pp, data_frame)

            if common.isInteger(self.plot_param.unique_item_count):
                # must be apply in this order: color -> shape

                #"""
                #color_count	= math.ceil(self.plot_param.unique_item_count * 0.3)
                #vector_size	= self.plot_param.unique_item_count
                pp = pp + ggplot2.scale_colour_manual(
                    values=self.plot_param.getPartiallyColoredVector(),
                    guide=guide_legend)

                shape_vector = getShapeVector(
                    self.plot_param.unique_item_count)
                pp = pp + ggplot2.scale_shape_manual(
                    values=shape_vector, guide=guide_legend)

            return self.save(pp, data_frame)

    class GGPlotStackBarPlotter(GGPlotTimelinePlotter):

        def __init__(self, plot_param, dry_run):
            super(GGPlotStackBarPlotter, self).__init__(plot_param, dry_run)

        def getPlotType(self):
            return PlotType.STACK_GRAPH

        def getWidthPixel(self, data_frame):
            return self.plot_param.getImageWidthPixel(data_frame)

        def plot(self, data_frame):
            #from rpy2.robjects.packages import importr
            #base = importr('base')
            #group	= base.gl(labels = ["Ctl","Trt"])
            #robj.globalenv["group"]	= group

            # print [column.rclass[0] for column in data_frame]
            common.verifyVarType(isRDataFrame, data_frame)
            self.showPlotInfo(data_frame)

            # data_frame.printDataframe()

            # required for desc()
            robj.r("library('plyr')")

            aes_string_args = self.plot_param.getAesStringArgs()
            if self.plot_param.stack_style == PlotType.Stack.BAR:
                if common.isNotEmptyString(self.plot_param.order):
                    aes_string_args["order"] = "desc(%s)" % (
                        self.plot_param.order)
            theme_args = dict(self._DEFAULT_THEME)
            # if common.isNotEmptyString(self.legend_title):
            #	theme_args["legend.title"]	= self.legend_title

            geom_bar_args = {
                "size"	: 0.25,
                "stat"	: "identity",
                "alpha"	: 0.6,
            }

            guide_legend = self.plot_param.getGuideLegend(
                self.getWidthPixel(data_frame), data_frame)

            group = aes_string_args.get("group")
            if group:
                unique_group_list = data_frame.getUniqueList(group)
                logger.debug(
                    "group: %d (%s)" % (len(unique_group_list), str(unique_group_list)))

            pp = self._ggplotBasic(
                data_frame, aes_string_args, theme_args, guide_legend)

            color_count = math.ceil(self.plot_param.unique_item_count * 0.3)
            vector_size = self.plot_param.unique_item_count

            if self.plot_param.stack_style == PlotType.Stack.BAR:
                geom_bar_args["alpha"] = 1.0
                pp = pp + ggplot2.geom_bar(**geom_bar_args)
                if common.isInteger(self.plot_param.unique_item_count):
                    pp = pp + ggplot2.scale_fill_manual(
                        values=getGrayColorVector(color_count, vector_size),
                        guide=guide_legend)
                    pp = pp + ggplot2.scale_colour_manual(
                        #values=getColorVector(color_count, vector_size),
                        values=self.plot_param.getPartiallyColoredVector(),
                        guide=guide_legend)

                    """
					linetype_range	= 6
					loop_count	= int(math.ceil(self.unique_item_count / float(linetype_range)))
					linetype_vector	= robj.IntVector(range(1, linetype_range + 1) * loop_count)
					logger.debug("linetype values: " + str(linetype_vector))
					pp	= pp + ggplot2.scale_linetype_manual(values=linetype_vector)
					"""
            elif self.plot_param.stack_style == PlotType.Stack.AREA:
                geom_bar_args["color"] = ggplot2.element_blank()
                pp = pp + ggplot2.geom_area(**geom_bar_args)
                #pp	= pp + ggplot2.geom_line(position="stack", size=0.1, colour="black")
                pp = pp + ggplot2.geom_line(position="stack", size=0.1)
                #color_vector	= getGrayColorVector(vector_size, vector_size * 2)
                color_vector = getGrayColorVector(vector_size, vector_size)
                pp = pp + ggplot2.scale_colour_manual(
                    values=color_vector, guide=guide_legend)
            else:
                pass

            common.debugDict(geom_bar_args, locals())

            pp = self.plotTimeRangeVerticalLine(pp, data_frame)

            if not common.isEmptyListOrTuple(self.plot_param.ylim):
                pp = pp + ggplot2.coord_cartesian(
                    ylim=robj.FloatVector(self.plot_param.ylim))

            return self.save(pp, data_frame)

    class GGPlotRectPlotter(GGPlotBasePlotter):

        def __init__(self, plot_param, dry_run):
            super(GGPlotRectPlotter, self).__init__(plot_param, dry_run)

        def getPlotType(self):
            return PlotType.RECT

        def getWidthPixel(self, data_frame):
            return self.plot_param.getImageWidthPixel(data_frame)

        def plot(self, data_frame):
            common.verifyVarType(isRDataFrame, data_frame)

            self.showPlotInfo(data_frame)

            aes_string_args = self.plot_param.getAesStringArgs()
            theme_args = dict(self._DEFAULT_THEME)

            aes_string_args["xmin"] = self.plot_param.xmin_vector
            aes_string_args["xmax"] = self.plot_param.xmax_vector
            aes_string_args["ymin"] = self.plot_param.ymin_vector
            aes_string_args["ymax"] = self.plot_param.ymax_vector
            common.debugDict(aes_string_args, locals())
            width_pixel = self.plot_param.getImageWidthPixel(data_frame)
            guide_legend = self.plot_param.getGuideLegend(
                width_pixel, data_frame)

            pp = self._ggplotBasic(
                data_frame, aes_string_args, theme_args, guide_legend)
            #pp	= pp + ggplot2.scale_y_continuous(name="y")
            #pp	= pp + ggplot2.scale_x_continuous(name="x")

            geom_rect_args = {
                "size"		: 0.05,
                "colour"	: "black",
                "alpha"		: 0.7,
                #"mapping":	ggplot2.aes_string(**aes_string_args)
                #"mapping":	aes_string_args,
            }
            pp = pp + ggplot2.geom_rect(**geom_rect_args)

            return self.save(pp, data_frame)

    class GGPlotHistgramPlotter(GGPlotBasePlotter):

        def __init__(self, plot_param, dry_run):
            GGPlotBasePlotter.__init__(self, plot_param, dry_run)

        def getPlotType(self):
            return PlotType.HISTGRAM

        def getWidthPixel(self, data_frame):
            return self.plot_param.getImageWidthPixel(data_frame)

        def plot(self, data_frame):
            common.verifyVarType(isRDataFrame, data_frame)

            self.showPlotInfo(data_frame)

            aes_string_args = self._getAesStringArgs()
            theme_args = dict(self._DEFAULT_THEME)
            #theme_args["panel.grid.minor.y"]	= ggplot2.element_blank()

            y_data_field = self.plot_param.getYDataField()
            x_data_field = self.plot_param.getXDataField()

            aes_string_args["x"] = x_data_field
            aes_string_args["y"] = y_data_field
            geom_hist_args = {
                "x"		: x_data_field,
                "stat"	: "identity",
            }
            if aes_string_args.get("size") is None:
                geom_hist_args["size"] = 1.5

            common.debugDict(geom_hist_args, locals())

            width_pixel = self._getImageWidthPixel(data_frame)
            guide_legend = self._getGuideLegend(width_pixel, data_frame)

            pp = self._ggplotBasic(
                data_frame, aes_string_args, theme_args, guide_legend)
            pp = pp + ggplot2.geom_bar(**geom_hist_args)
            #pp	= pp + ggplot2.geom_histogram(**geom_point_args)

            return self.save(pp, data_frame)
