# encoding: utf-8

'''
@author: vmmhypervisor
'''

import os
import re

import golib
import golib.common as common
import golib.sqlite as sql
import golib.sysstat as sysstat
import golib.gfile as gfile
from golib.logger import logger


class ActivityOption:
    SKIP = "skip"
    MAKE = "make"
    OVERWRITE = "overwrite"
    CLEAN = "clean"
    __CHOICE_LIST = [SKIP, MAKE, OVERWRITE, CLEAN]

    @classmethod
    def addSarReportArgument(cls, parser, group=None):
        choice_list = cls.__CHOICE_LIST

        if group is None:
            group = parser.add_argument_group("Sysstat")

        group.add_argument(
            "--sar-report",
            dest="sar_report_make_option",
            type=str, choices=choice_list, default=choice_list[0],
            help="default=%(default)s")

        return group

    @classmethod
    def addPfmReportArgument(cls, parser, default_option=MAKE):
        choice_list = cls.__CHOICE_LIST

        group = parser.add_argument_group("PFM")
        group.add_argument(
            "--pfm-report",
            dest="pfm_report_make_option",
            type=str, choices=choice_list, default=choice_list[0],
            help=" default=%(default)s")

    @classmethod
    def addSarFileNamePatternArgument(cls, parser, group=None):
        if group is None:
            group = parser.add_argument_group("Input File Name Pattern")

        group.add_argument(
            "--sar-text-pattern",
            default=SarStatsMakeProperty.Default.Pattern.SAR_TEXT_FILE,
            help="regular expression of sar text file name (default=%(default)s)")
        group.add_argument(
            "--sar-bin-pattern",
            default=SarStatsMakeProperty.Default.Pattern.SAR_BIN_FILE,
            help="regular expression of sar binary file name (default=%(default)s)")

        return group


class SarStatsMakeProperty(common.BaseObject):

    class Default:

        class Pattern:
            SAR_TEXT_FILE = "^%s.*%s$" % (
                sysstat.RAW_SAR_FILE_NAME_PREFIX, re.escape(".txt"))
            SAR_BIN_FILE = "^%s.*%s$" % (
                sysstat.RAW_SAR_FILE_NAME_PREFIX,
                re.escape(sysstat.EXT_SAR_BIN_DATA))

            CPU_INFO = "^cpuinfo"
            INTERRUPTS = "^interrupts"
            DISK_STATS = "^diskstats"
            DLNKMGR = "^dlnkmgr_view_drv"
            LUMAPLIST = "^lumaplist"
            FSLIST = "^fslist_c.*\.txt$"

        SAR_REPORT_OUTPUT_DIR_NAME_PREFIX = "sar-report"

    def __init__(self):
        super(SarStatsMakeProperty, self).__init__()

        self.make_option = ActivityOption.MAKE

        self.search_dir_path = None
        self.output_base_dir_path = None
        self.out_dir_name_suffix = ""
        self.__out_dir_name_prefix = self.Default.SAR_REPORT_OUTPUT_DIR_NAME_PREFIX

        self.datetime_range = None
        self.discard_rate = None

        self.sar_file_path = None
        self.sar_txt_file_name_pattern = self.Default.Pattern.SAR_TEXT_FILE
        self.sar_bin_file_name_pattern = self.Default.Pattern.SAR_BIN_FILE

        self.re_cpuinfo = self.Default.Pattern.CPU_INFO
        self.re_interrupts = self.Default.Pattern.INTERRUPTS
        self.re_diskstats = self.Default.Pattern.DISK_STATS
        self.re_dlnkmgr = self.Default.Pattern.DLNKMGR
        self.re_lumaplist = self.Default.Pattern.LUMAPLIST
        self.re_fslist = self.Default.Pattern.FSLIST

    def getOutputDirectoryPath(self):
        dir_name_item_list = []

        if common.isEmptyString(self.output_base_dir_path):
            if common.isNotEmptyString(self.search_dir_path):
                self.output_base_dir_path = self.search_dir_path
            else:
                self.output_base_dir_path = "."

        if common.isNotEmptyString(self.__out_dir_name_prefix):
            dir_name_item_list.append(self.__out_dir_name_prefix)

        if self.datetime_range is not None:
            self.datetime_range.time_format = golib.gtime.Format.ISO8601.Basic.DATETIME
            text = self.datetime_range.getStartTimeText()
            if common.isNotEmptyString(text):
                dir_name_item_list.append(text)

        if common.isNotEmptyString(self.out_dir_name_suffix):
            dir_name_item_list.append(self.out_dir_name_suffix)

        output_dir_path = os.path.join(
            self.output_base_dir_path, "_".join(dir_name_item_list))

        return os.path.realpath(output_dir_path)
        # return common.getRelativePath(file_output_dir_path, os.getcwd())


class SarReportMaker(common.BaseObject):

    class FilePath:
        RAW_SAR_DB = "raw_sar_db"
        SAR_STATS_DB = "sar_stats_db"
        SAR_STATS_TEXT = "sar_stats_text"

    def __init__(self, sys_wrapper, make_prop, base_option_list):
        """
        make_prop: SarStatsMakeProperty
        """

        self.__sys_wrapper = sys_wrapper
        self.setProperty(make_prop)
        self.__base_option_list = base_option_list

    def setProperty(self, make_prop):
        self.__make_prop = make_prop

    def cleanReportFiles(self):
        make_prop = self.__make_prop
        make_prop.debug("clean sar report files")

        re_pattern = "^%s" % (
            SarStatsMakeProperty.Default.SAR_REPORT_OUTPUT_DIR_NAME_PREFIX)

        search_root_dir_path = make_prop.output_base_dir_path
        remove_dir_path_list = gfile.findFileAll(
            search_root_dir_path, os.path.isdir, re_pattern)
        if common.isEmptyString(remove_dir_path_list):
            logger.warn("remove directory not found: " + re_pattern)
            return
        for remove_dir_path in remove_dir_path_list:
            gfile.FileManager.removeDirectory(remove_dir_path)

    def findSarFile(self):
        """
        sarファイルを探索。
        text形式のsarファイルを優先。

        :return: sar file path
        :raise:
                - gfile.FileNotFoundError
        """

        logger.debug("findSarFile")

        prop = self.__make_prop

        if common.isNotEmptyString(prop.sar_file_path):
            return prop.sar_file_path

        # find sar text file---
        sar_file_path_list = gfile.findFileAll(
            prop.search_dir_path, os.path.isfile,
            prop.sar_txt_file_name_pattern)
        for sar_file_path in sar_file_path_list:
            if not sysstat.Sar.isValidTextDataFile(sar_file_path):
                continue
            sar_file_path = gfile.getRelativePath(sar_file_path, os.getcwd())
            logger.debug("sar text file found: " + sar_file_path)
            return sar_file_path

        # find sar binary file ---
        sar_file_path_list = gfile.findFileAll(
            prop.search_dir_path, os.path.isfile,
            prop.sar_bin_file_name_pattern)
        for sar_file_path in sar_file_path_list:
            try:
                if not sysstat.Sar.isValidBinDataFile(sar_file_path):
                    continue
            except common.NotInstallError, e:
                logger.error(e)
                break
            sar_file_path = gfile.getRelativePath(sar_file_path, os.getcwd())
            logger.debug("sar binary file found: " + sar_file_path)
            return sar_file_path

        raise gfile.FileNotFoundError("sar file not found in %s: %s, %s" % (
            prop.search_dir_path,
            prop.sar_txt_file_name_pattern, prop.sar_bin_file_name_pattern))

    def __makeWrapper(self, output_path, command):
        """
        raise:
                RuntimeError: コマンドの返り値が0以外の場合
        """

        prop = self.__make_prop

        if prop.make_option == ActivityOption.SKIP:
            logger.debug("skip: " + output_path)
            return 0

        if prop.make_option == ActivityOption.OVERWRITE:
            return self.__sys_wrapper.run(command)

        try:
            gfile.check_file_existence(output_path)
        except:
            pass
        else:
            logger.info("skip make: " + output_path + " already exists")
            return 0

        return_value = self.__sys_wrapper.run(command)
        if return_value != 0:
            raise RuntimeError("%s: return-value=%d" % (command, return_value))

        return return_value

    def __convertSar(self, sar_file_path, output_path):
        prop = self.__make_prop

        option_list = (self.__base_option_list + [
            "-f", sar_file_path,
            prop.datetime_range.getOptionString(),
        ])
        if common.isNotEmptyString(output_path):
            option_list.extend(["-o", output_path])

        cpuinfo_path = gfile.findFile(prop.search_dir_path, prop.re_cpuinfo)
        if common.isNotEmptyString(cpuinfo_path):
            option_list.extend(["--cpuinfo", cpuinfo_path])

        interrupts_path = gfile.findFile(
            prop.search_dir_path, prop.re_interrupts)
        if common.isNotEmptyString(interrupts_path):
            option_list.extend(["--interrupts", interrupts_path])

        diskstats_path = gfile.findFile(
            prop.search_dir_path, prop.re_diskstats)
        if common.isNotEmptyString(diskstats_path):
            option_list.extend(["--diskstats", diskstats_path])

        dlnkmgr_path = gfile.findFile(prop.search_dir_path, prop.re_dlnkmgr)
        if common.isNotEmptyString(dlnkmgr_path):
            option_list.extend(["--dlnkmgr", dlnkmgr_path])

            lumaplist_path = gfile.findFile(
                prop.search_dir_path, prop.re_lumaplist)
            if common.isNotEmptyString(lumaplist_path):
                option_list.extend(["--lumaplist", lumaplist_path])

                fslist_path = gfile.findFile(
                    prop.search_dir_path, prop.re_fslist)
                if common.isNotEmptyString(fslist_path):
                    option_list.extend(["--fslist", fslist_path])

        base_command = "convert_sar"
        command = base_command + " " + " ".join(option_list)
        return self.__makeWrapper(output_path, command)

    def __makeSarStats(
            self, raw_sar_db_path, output_path):

        make_prop = self.__make_prop

        option_list = (
            self.__base_option_list +
            [
                "-f", raw_sar_db_path,
                "-A",
            ])

        if common.isNotEmptyString(output_path):
            option_list.extend(["-o", output_path])

        if common.isFloat(make_prop.discard_rate):
            option_list.extend(["--strip", str(make_prop.discard_rate)])

        base_command = "make_sar_statistics"
        command = base_command + " " + " ".join(option_list)

        return self.__makeWrapper(output_path, command)

    def makeSarReport(self):
        """
        :raise:
                - NotInstallError
        """

        SAR_STATISTICS_SUFIX = "stats"

        prop = self.__make_prop
        prop.debug("makeSarReport")

        if prop.make_option == ActivityOption.SKIP:
            return {}

        """
        if prop.make_option == ActivityOption.CLEAN:
            self.cleanReportFiles()
            return
        """

        common.checkInstallCommand([
            "convert_sar", "make_sar_statistics", "show_sar_statistics"
        ])

        sar_file_path = self.findSarFile()
        base_filename = common.getListItem(
            os.path.splitext(os.path.basename(sar_file_path)), 0)
        output_dir_path = prop.getOutputDirectoryPath()

        gfile.FileManager.makeDirectory(output_dir_path)

        raw_sar_db_path = os.path.join(
            output_dir_path, base_filename + sql.SQLITE_EXTENSION)
        stats_file_base_name = "sar_" + SAR_STATISTICS_SUFIX
        sar_stats_db_path = os.path.join(
            output_dir_path, stats_file_base_name + sql.SQLITE_EXTENSION)

        self.__convertSar(sar_file_path, output_path=raw_sar_db_path)
        self.__makeSarStats(raw_sar_db_path, output_path=sar_stats_db_path)

        # output sar summary ---
        sar_stats_txt_path = os.path.join(
            output_dir_path, stats_file_base_name + ".txt")

        output_path = sar_stats_txt_path
        option_list = (
            self.__base_option_list +
            [
                "-f", sar_stats_db_path,
                "-A",
                "--verbose",
            ])

        if prop.make_option == golib.option.MakeOption.OVERWRITE:
            option_list.extend(["--overwrite"])

        show_stats_script_name = "show_sar_statistics"
        command = "%s %s > %s" % (
            show_stats_script_name, " ".join(option_list), output_path)
        self.__makeWrapper(output_path, command)

        # make html report ---
        try:
            option_list.extend([
                "--style", "html",
                "--output-dir", output_dir_path,
            ])
            command = "%s %s" % (show_stats_script_name, " ".join(option_list))
            if prop.make_option != ActivityOption.SKIP:
                self.__sys_wrapper.run(command)
        except:
            pass

        return {
            self.FilePath.RAW_SAR_DB		: raw_sar_db_path,
            self.FilePath.SAR_STATS_TEXT	: sar_stats_txt_path,
            self.FilePath.SAR_STATS_DB		: sar_stats_db_path,
        }


def getMaxTransferNetworkInterface(con_stats):
    import golib.statistics as stats

    where_list = [
        sql.SqlQuery.makeWhereNotIn(
            "IFACE", [common.KEY_ALL, common.KEY_TOTAL, common.KEY_AVERAGE]),
        sql.SqlQuery.make_where(
            common.AN_GeneralKey, sysstat.FN_TotalKbPerSec),
    ]

    nic_list = con_stats.getValueList(
        select="IFACE",
        table=sysstat.TN_NetworkDev,
        where=" AND ".join(where_list),
        extra="ORDER BY %s DESC" % (sql.SqlQuery.to_attr_str(stats.MEAN)))
    for nic in nic_list:
        if re.search("^agr|^xgbe|^eth|^mng", nic):
            return nic

    return None


def connectSarStatsDatabase(database_path, options):
    """
    sar統計データベースへのコネクションを返す。
    """

    return sql.connect_sqlite_database(
        database_path, "r",
        sysstat.DN_SarStatistics, sysstat.DV_SarStatistics, options)


def getSarStatistics(con, table_name, where_query):
    import golib.statistics as stats

    null_result = ({}, {})

    if con is None:
        logger.warn("null database connection")
        return null_result

    if not con.hasTable(table_name):
        logger.error(
            "'%s' table not found in %s" % (table_name, con.database_path))
        return null_result

    select_list = [
        sql.SqlQuery.to_attr_str(common.AN_GeneralKey),
        sql.SqlQuery.to_attr_str(sysstat.SUMARRY_STATISTICS_TYPE),
        sql.SqlQuery.to_attr_str(stats.RSD),
    ]

    query_result = con.execute_select(
        select=",".join(select_list),
        table=table_name,
        where=where_query)
    if query_result is None:
        logger.error("empty result")
        return null_result

    dict_Key_Util = {}
    dict_Key_Rsd = {}
    for key, util, rsd in query_result.fetchall():
        dict_Key_Util[key] = util
        dict_Key_Rsd[key] = rsd

    return (dict_Key_Util, dict_Key_Rsd)


"""
def getLoadAverageEx(con):
    table_name = sysstat.TN_Queue
    where_query = sql.SqlQuery.make_where(common.AN_GeneralKey, "ldavg-1")

    return getSarStatistics(con, table_name, where_query)


def getSarBlockIoStatistics(con):
    table_name = sysstat.TN_BlockIO

    where_list = [
        sql.SqlQuery.make_where(common.AN_GeneralKey, sysstat.FN_ReadKbPerSec),
        sql.SqlQuery.make_where(common.AN_GeneralKey, sysstat.FN_WriteKbPerSec),
    ]
    where_query = " OR ".join(where_list)

    return getSarStatistics(con, table_name, where_query)


def getDiskStatistics(con):
    import golib.sysstat

    if con is None:
        return ({}, {})

    table_name = golib.sysstat.getDiskTableName(
        sysstat.TN_DiskIO, sysstat.FN_FileSystem)

    if not con.hasTable(table_name):
        logger.debug(
            "'%s' table not found in %s" % (table_name, con.database_path))
        return ({}, {})

    where_list = [
        sql.SqlQuery.make_where(common.AN_GeneralKey, "tps"),
        sql.SqlQuery.make_where(common.AN_GeneralKey, "await"),
        sql.SqlQuery.make_where(common.AN_GeneralKey, "svctm"),
    ]
    where_query = "%s AND (%s)" % (sql.SqlQuery.make_where(
        sysstat.FN_FileSystem, common.KEY_AVERAGE), " OR ".join(where_list))

    return getSarStatistics(con, table_name, where_query)
#"""


def getSarNetworkStatistics(con, interface):
    table_name = sysstat.TN_NetworkDev

    where_list = [
        sql.SqlQuery.make_where(common.AN_GeneralKey, "rxkB/s"),
        sql.SqlQuery.make_where(common.AN_GeneralKey, "txkB/s"),
    ]
    where_query = "%s AND (%s)" % (
        sql.SqlQuery.make_where("IFACE", interface), " OR ".join(where_list))

    return getSarStatistics(con, table_name, where_query)


def getSarCpuUtil(con):
    table_name = sysstat.TN_CpuUtil

    where_list = [
        sql.SqlQuery.make_where(common.AN_GeneralKey, sysstat.FN_SarCpuUtil),
        sql.SqlQuery.make_where(common.AN_GeneralKey, "%user"),
        sql.SqlQuery.make_where(common.AN_GeneralKey, "%system"),
        sql.SqlQuery.make_where(common.AN_GeneralKey, "%iowait"),
    ]
    where_query = "%s AND (%s)" % (sql.SqlQuery.make_where(
        sysstat.FN_Cpu, common.KEY_AVERAGE), " OR ".join(where_list))

    return getSarStatistics(con, table_name, where_query)
